# VirtIO驱动框架接口

[返回主页](./virtio.md)

本节主要列举可以在VirtIO前端驱动中使用的通用VirtIO驱动框架接口。当需要实现新的VirtIO设备时，必须使用这些接口实现的VirtIO协议。

## 函数方法

### virtio.h

```C
/* Common Config */
void VirtioCommonConfigGet(NX_Device *device, VirtioCommonConfigItem item, void *value);
void VirtioCommonConfigSet(NX_Device *device, VirtioCommonConfigItem item, void *value);
/* Device Config */
void *VirtioGetDeviceConfig(NX_Device *device);
void VirtioReadDeviceConfig(NX_Device *device, NX_Offset offset, void *buf, NX_Size len);
/* Notification Config, VIRTIO_F_NOTIFICATION_DATA should not be negotiated */
// From Device, notifying Driver about Used
void VirtioWaitNotification(NX_Device *device, NX_U16 queue_index, NX_U32 desc_id);
// From Driver, notifying Device about Avail
void VirtioSendNotification(NX_Device *device, NX_U16 queue_index, NX_U32 desc_id);
/* Generic Interrupt Service Routine */
NX_Error VirtioISR(VirtioISRType type, NX_Device *device);
NX_Error VirtioSetupISR(NX_Device *device);
NX_Error VirtioRemoveISR(NX_Device *device);
/* Generic Initialization */
NX_Error VirtioDeviceInitialize(NX_Device *device, NX_U32 *driverFeatures, NX_Size driverFeaturesCount);
NX_Error VirtioDeviceDriverOK(NX_Device *device);
/* Generic Cleanup */
NX_Error VirtioDeviceReset(NX_Device *device);
void VirtioCleanUpDevice(NX_Device *device);

NX_INLINE void *VirtioVirt2Phy(void *buf);
NX_INLINE NX_Addr VirtioVirtualAddressAllocate(NX_Addr phyAddr);
```

### virtio_queue.h

```C
void VirtioQueueInit(NX_Device *device, NX_U16 queue_index, NX_U16 ring_size);
void VirtioQueueFree(NX_Device *device);
void VirtioQueueAlloc(NX_Device *device, int queue_num);
void VirtioQueueDestroy(NX_Device *device, NX_U16 queue_index);
VirtioQueue *VirtioQueueGet(NX_Device *device, NX_U16 queue_index);

NX_U32 VirtioQueueDescAlloc(VirtioQueue *queue);
NX_Error VirtioQueueDescAllocMany(VirtioQueue *queue, int num, NX_U32 *allocated_ids);
void VirtioQueueDescFree(VirtioQueue *queue, NX_U32 id);
void VirtioQueueDescFreeMany(VirtioQueue *queue, NX_U32 id);
VirtioQueueDesc *VirtioQueueDescGet(VirtioQueue *queue, NX_U32 id);

void VirtioQueueDescFill(VirtioQueue *queue, NX_U32 desc_id, NX_U64 addr, NX_U32 len, NX_U16 flags, NX_U16 next);
void VirtioQueuePutIntoAvailRing(VirtioQueue *queue, NX_U32 desc_id);

void VirtioQueueRegisterNotifier(VirtioQueue *queue, VirtioQueueCompletedHandler handler);
void VirtioQueueUnregisterNotifier(VirtioQueue *queue, VirtioQueueCompletedHandler handler);
```

## 宏定义

### virtio.h
```C
/* VIRTIO Generic Definitions */
#define VIRTIO_STATUS_RESET         0 // Write 0 to Status reset the device

#define VIRTIO_STATUS_ACKNOWLEDGE   1
#define VIRTIO_STATUS_DRIVER        2
#define VIRTIO_STATUS_DRIVER_OK     4
#define VIRTIO_STATUS_FEATURES_OK   8
#define VIRTIO_STATUS_NEEDS_RESET   64
#define VIRTIO_STATUS_FAILED        128

// Device-independent features bits (modern)
// 0 ~ 31
#define VIRTIO_FEATURE_INDIRECT_DESC        (1<<28)
#define VIRTIO_FEATURE_EVENT_IDX            (1<<29)
// 32 ~ 63
#define VIRTIO_FEATURE_VERSION_1            (1<<(32 - 32))
#define VIRTIO_FEATURE_ACCESS_PLATFORM      (1<<(33 - 32))
#define VIRTIO_FEATURE_RING_PACKED          (1<<(34 - 32))
#define VIRTIO_FEATURE_IN_ORDER             (1<<(35 - 32))
#define VIRTIO_FEATURE_ORDER_PLATFORM       (1<<(36 - 32))
#define VIRTIO_FEATURE_SR_IOV               (1<<(37 - 32))
#define VIRTIO_FEATURE_NOTIFICATION_DATA    (1<<(38 - 32))
#define VIRTIO_FEATURE_NOTIF_CONFIG_DATA    (1<<(39 - 32))
#define VIRTIO_FEATURE_RING_RESET           (1<<(40 - 32))

// Device-independent features bits (legacy)
#define VIRTIO_FEATURE_NOTIFY_ON_EMPTY      (1<<24)
#define VIRTIO_FEATURE_ANY_LAYOUT           (1<<27)
#define VIRTIO_FEATURE_UNUSED               (1<<30)

#define VIRTIO_PAGE_SIZE            4096

#define VirtioGetExtension(device) ((VirtioExtension*)(device)->extension)
#define VirtioGetBusExtension(device, type) ((type*)(VirtioGetExtension(device)->busExt))
#define VirtioGetDevExtension(device, type) ((type*)(VirtioGetExtension(device)->devExt))

// For VIRTIO_COMMON_CFG_QUEUE_ENABLE
#define VIRTIO_COMMON_CFG_QUEUE_ENABLE_ENABLE 1
#define VIRTIO_COMMON_CFG_QUEUE_ENABLE_DISABLE 0

// For VirtioVirtualAddressAllocate
#define VIRTIO_VADDR_MASK   0xF0000000;
```

### virtio_queue.h

```C
#define VIRTIO_QUEUE_DESC_FLAG_NEXT     1   // marks a buffer as continuing via the next field
#define VIRTIO_QUEUE_DESC_FLAG_WRITE    2   // marks a buffer as write-only
#define VIRTIO_QUEUE_DESC_FLAG_INDIRECT 4   // the buffer contains a list of buffer descriptors

#define VIRTIO_QUEUE_USED_FLAG_NO_NOTIFY        1   // no notify when add a buffer (unreliable)
#define VIRTIO_QUEUE_AVAIL_FLAG_NO_INTERRUPT    1   // no interrupt then consume a buffer (unreliable)

#define VIRTIO_QUEUE_DESC_TOTAL_SIZE(ring_size)     (sizeof(VirtioQueueDesc) * (ring_size))
#define VIRTIO_QUEUE_AVAIL_TOTAL_SIZE(ring_size)    (sizeof(VirtioQueueAvail) + (ring_size) * sizeof(NX_U16))
#define VIRTIO_QUEUE_USED_TOTAL_SIZE(ring_size)     (sizeof(VirtioQueueUsed) + (ring_size) * sizeof(VirtioQueueUsedElement))

#define VIRTIO_QUEUE_DESC_INVALID_ID (0xFFFFFFFF)
```

## 结构体定义

### virtio.h

```C
// VirtioExtension
typedef void                   (*VirtioConfigChangeHandler)(NX_Device *device);

struct VirtioExtension
{
    VirtioQueue               *queue;
    int                       queue_num;
    VirtioBusType             busType;
    void                      *busExt;
    void                      *devExt;
    NX_Device                 *device; // For reverse lookup
    NX_U32                    flags;
    NX_List                   list;
    NX_IRQ_DelayWork          *irqWorkQueue;
    NX_IRQ_DelayWork          *irqWorkConfig;
    VirtioConfigChangeHandler configChangeNotifier;
};
typedef struct VirtioExtension VirtioExtension;

// VirtioDriverTableEntity

typedef NX_Error                       (*VirtioDriverInitHandler)(NX_Driver *driver, NX_Device *device);
typedef void                           (*VirtioDriverCleanUpHandler)(NX_Device *device);
struct VirtioDriverTableEntity
{
    const VirtioDeviceId       deviceId;                    /* Static, Driver provide */
    const char                 *driverName;                 /* Static, Driver provide */
    const char                 *deviceNameTemplate;         /* Static, Driver provide */
    const NX_DeviceType        deviceType;                  /* Static, Driver provide */
    NX_DriverOps               *driverOps;                  /* Static, Driver provide */
    VirtioDriverInitHandler    initHandler;                 /* Static, Driver provide */
    VirtioDriverCleanUpHandler cleanUpHandler;              /* Static, Driver provide */
    NX_Driver                  *driver;                     /* Runtime */
    int                        deviceCount;                 /* Runtime */
    NX_List                    deviceListHead;              /* Runtime */
};
typedef struct VirtioDriverTableEntity VirtioDriverTableEntity;

// VirtioFeature

struct VirtioFeature
{
    NX_U32 selector;
    NX_U32 features;
};
typedef struct VirtioFeature           VirtioFeature;

```

### virtio_queue.h

```C
/* Virtqueue descriptors: 16 bytes.
 * These can chain together via "next". */
struct VirtioQueueDesc
{
    /* Address (guest-physical). */
    NX_U64 addr;
    /* Length. */
    NX_U32 len;
    /* The flags as indicated above. */
    NX_U16 flags;
    /* We chain unused descriptors via this, too */
    NX_U16 next;
};
typedef struct VirtioQueueDesc        VirtioQueueDesc;

struct VirtioQueueAvail
{
    NX_U16 flags;
    NX_U16 idx;
    NX_U16 ring[];
    /* Only if VIRTIO_F_EVENT_IDX: NX_U16 used_event; */
};
typedef struct VirtioQueueAvail       VirtioQueueAvail;

/* NX_U32 is used here for ids for padding reasons. */
struct VirtioQueueUsedElement
{
    /* Index of start of used descriptor chain. */
    NX_U32 id;
    /* Total length of the descriptor chain which was written to. */
    NX_U32 len;
};
typedef struct VirtioQueueUsedElement VirtioQueueUsedElement;

struct VirtioQueueUsed
{
    NX_U16                        flags;
    NX_U16                        idx;
    struct VirtioQueueUsedElement ring[];
    /* Only if VIRTIO_F_EVENT_IDX: NX_U16 avail_event; */
};
typedef struct VirtioQueueUsed        VirtioQueueUsed;

struct VirtioQueueTrace
{
    /*
     * For synchronous waiting
     */
    NX_Bool      waiting;    // Driver is expecting this idx
    NX_Bool      ok;         // Interrupt has been made
    NX_Semaphore lock;
    /*
     * For asynchronous completed notifying
     */
    NX_Bool      needNotify;
};
typedef struct VirtioQueueTrace       VirtioQueueTrace;

typedef struct VirtioQueue VirtioQueue;

typedef NX_Error(*VirtioQueueCompletedHandler)(NX_Device *device, VirtioQueue *queue, NX_U32 idx);

struct VirtioQueue
{
    NX_U16       queue_index;
    unsigned int ring_size; // be power of 2
    NX_U32       used_idx;

    VirtioQueueDesc  *desc; // alignment of 16
    VirtioQueueAvail *avail;// alignment of 2
    VirtioQueueUsed  *used; // alignment of 4

    NX_Bool          *free; // consider using bitset to reduce memory usage
    unsigned int     free_count;
    VirtioQueueTrace *trace;

    VirtioQueueCompletedHandler notifier;

    NX_Mutex freeLock;
    NX_Mutex availRingLock;
    NX_Spin  usedRingLock;
};

```